@@ -1,13 +1,13 @@
 ;;; marginalia.el --- Enrich existing commands with completion annotations -*- lexical-binding: t -*-
 
-;; Copyright (C) 2021-2023 Free Software Foundation, Inc.
+;; Copyright (C) 2021-2024 Free Software Foundation, Inc.
 
 ;; Author: Omar Antolín Camarena <omar@matem.unam.mx>, Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Omar Antolín Camarena <omar@matem.unam.mx>, Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2020
-;; Version: 1.3
-;; Package-Requires: ((emacs "27.1") (compat "29.1.4.0"))
-;; Homepage: https://github.com/minad/marginalia
+;; Version: 1.8
+;; Package-Requires: ((emacs "28.1") (compat "30"))
+;; URL: https://github.com/minad/marginalia
 ;; Keywords: docs, help, matching, completion
 
 ;; This file is part of GNU Emacs.
@@ -41,7 +41,7 @@
 (defgroup marginalia nil
   "Enrich existing commands with completion annotations."
   :link '(info-link :tag "Info Manual" "(marginalia)")
-  :link '(url-link :tag "Homepage" "https://github.com/minad/marginalia")
+  :link '(url-link :tag "Website" "https://github.com/minad/marginalia")
   :link '(emacs-library-link :tag "Library Source" "marginalia.el")
   :group 'help
   :group 'docs
@@ -75,32 +75,42 @@ Set to `most-positive-fixnum' to always use a relative age, or 0 to never show
 a relative age."
   :type 'natnum)
 
+(defcustom marginalia-remote-file-regexps
+  '("\\`/\\([^/|:]+\\):") ;; Tramp path
+  "List of remote file regexps where the files should not be annotated.
+
+The first match group is displayed instead of the detailed file
+attribute information.  For Tramp paths, the protocol is
+displayed instead."
+  :type '(repeat regexp))
+
 (defcustom marginalia-annotator-registry
   (mapcar
    (lambda (x) (append x '(builtin none)))
-   '((command marginalia-annotate-command marginalia-annotate-binding)
-     (embark-keybinding marginalia-annotate-embark-keybinding)
-     (customize-group marginalia-annotate-customize-group)
-     (variable marginalia-annotate-variable)
-     (function marginalia-annotate-function)
-     (face marginalia-annotate-face)
-     (color marginalia-annotate-color)
-     (unicode-name marginalia-annotate-char)
-     (minor-mode marginalia-annotate-minor-mode)
-     (symbol marginalia-annotate-symbol)
-     (environment-variable marginalia-annotate-environment-variable)
-     (input-method marginalia-annotate-input-method)
-     (coding-system marginalia-annotate-coding-system)
-     (charset marginalia-annotate-charset)
-     (package marginalia-annotate-package)
-     (imenu marginalia-annotate-imenu)
-     (bookmark marginalia-annotate-bookmark)
-     (file marginalia-annotate-file)
-     (project-file marginalia-annotate-project-file)
-     (buffer marginalia-annotate-buffer)
-     (library marginalia-annotate-library)
-     (tab marginalia-annotate-tab)
-     (multi-category marginalia-annotate-multi-category)))
+   `((command ,#'marginalia-annotate-command ,#'marginalia-annotate-binding)
+     (embark-keybinding ,#'marginalia-annotate-embark-keybinding)
+     (customize-group ,#'marginalia-annotate-customize-group)
+     (variable ,#'marginalia-annotate-variable)
+     (function ,#'marginalia-annotate-function)
+     (face ,#'marginalia-annotate-face)
+     (color ,#'marginalia-annotate-color)
+     (unicode-name ,#'marginalia-annotate-char)
+     (minor-mode ,#'marginalia-annotate-minor-mode)
+     (symbol ,#'marginalia-annotate-symbol)
+     (environment-variable ,#'marginalia-annotate-environment-variable)
+     (input-method ,#'marginalia-annotate-input-method)
+     (coding-system ,#'marginalia-annotate-coding-system)
+     (charset ,#'marginalia-annotate-charset)
+     (package ,#'marginalia-annotate-package)
+     (imenu ,#'marginalia-annotate-imenu)
+     (bookmark ,#'marginalia-annotate-bookmark)
+     (file ,#'marginalia-annotate-file)
+     (project-file ,#'marginalia-annotate-project-file)
+     (buffer ,#'marginalia-annotate-buffer)
+     (library ,#'marginalia-annotate-library)
+     (theme ,#'marginalia-annotate-theme)
+     (tab ,#'marginalia-annotate-tab)
+     (multi-category ,#'marginalia-annotate-multi-category)))
   "Annotator function registry.
 Associates completion categories with annotation functions.
 Each annotation function must return a string,
@@ -108,10 +118,10 @@ which is appended to the completion candidate."
   :type '(alist :key-type symbol :value-type (repeat symbol)))
 
 (defcustom marginalia-classifiers
-  '(marginalia-classify-by-command-name
-    marginalia-classify-original-category
-    marginalia-classify-by-prompt
-    marginalia-classify-symbol)
+  (list #'marginalia-classify-by-command-name
+        #'marginalia-classify-original-category
+        #'marginalia-classify-by-prompt
+        #'marginalia-classify-symbol)
   "List of functions to determine current completion category.
 Each function should take no arguments and return a symbol
 indicating the category, or nil to indicate it could not
@@ -126,7 +136,7 @@ determine it."
     ("\\<color\\>" . color)
     ("\\<face\\>" . face)
     ("\\<environment variable\\>" . environment-variable)
-    ("\\<function\\|hook to remove\\>" . function)
+    ("\\<function\\|\\(?:hook\\|advice\\) to remove\\>" . function)
     ("\\<variable\\>" . variable)
     ("\\<input method\\>" . input-method)
     ("\\<charset\\>" . charset)
@@ -134,7 +144,8 @@ determine it."
     ("\\<minor mode\\>" . minor-mode)
     ("\\<kill-ring\\>" . kill-ring)
     ("\\<tab by name\\>" . tab)
-    ("\\<library\\>" . library))
+    ("\\<library\\>" . library)
+    ("\\<theme\\>" . theme))
   "Associates regexps to match against minibuffer prompts with categories.
 The prompts are matched case-insensitively."
   :type '(alist :key-type regexp :value-type symbol))
@@ -290,11 +301,10 @@ The value of `this-command' is used as key for the lookup."
 
 ;;;; Pre-declarations for external packages
 
-(declare-function project-current "project")
+(declare-function bookmark-prop-get "bookmark")
 
-(declare-function bookmark-get-handler "bookmark")
-(declare-function bookmark-get-filename "bookmark")
-(declare-function bookmark-get-front-context-string "bookmark")
+(declare-function project-current "project")
+(declare-function project-root "project")
 
 (defvar package--builtins)
 (defvar package-archive-contents)
@@ -311,6 +321,10 @@ The value of `this-command' is used as key for the lookup."
 
 ;;;; Marginalia mode
 
+(defalias 'marginalia--orig-completion-metadata-get
+  (symbol-function (compat-function completion-metadata-get))
+  "Original `completion-metadata-get' function.")
+
 (defvar marginalia--pangram "Cwm fjord bank glyphs vext quiz.")
 
 (defvar marginalia--bookmark-type-transforms
@@ -355,8 +369,6 @@ for performance profiling of the annotators.")
 (defun marginalia--ellipsis ()
   "Return ellipsis."
   (with-memoization marginalia--ellipsis
-    ;; TODO Emacs 28 offers the function `truncate-string-ellipsis'.
-    ;; It must be backported in Compat.
     (cond
      ((bound-and-true-p truncate-string-ellipsis))
      ((char-displayable-p ?…) "…")
@@ -405,6 +417,15 @@ FACE is the name of the face, with which the field should be propertized."
                `((marginalia--field ,@(cdr field)))))
            fields))))
 
+(defmacro marginalia--in-minibuffer (&rest body)
+  "Run BODY inside minibuffer if minibuffer is active.
+Otherwise stay within current buffer."
+  (declare (indent 0))
+  `(with-current-buffer (if-let (win (active-minibuffer-window))
+                            (window-buffer win)
+                          (current-buffer))
+     ,@body))
+
 (defun marginalia--documentation (str)
   "Format documentation string STR."
   (when str
@@ -425,15 +446,15 @@ FACE is the name of the face, with which the field should be propertized."
     (fun fun)))
 
 (defun marginalia-annotate-multi-category (cand)
-  "Annotate multi-category CAND with the buffer class."
+  "Annotate multi-category CAND, dispatching to the appropriate annotator."
   (if-let ((multi (get-text-property 0 'multi-category cand))
            (annotate (marginalia--annotator (car multi))))
       ;; Use the Marginalia annotator corresponding to the multi category.
       (funcall annotate (cdr multi))
-    ;; Apply the original annotation function on the original candidate, if
-    ;; there is one.  NOTE: Use `alist-get' instead of `completion-metadata-get'
-    ;; to bypass our `marginalia--completion-metadata-get' advice!
-    (when-let (annotate (alist-get 'annotation-function marginalia--metadata))
+    ;; Apply the original annotation function on the original candidate. Bypass
+    ;; our `marginalia--completion-metadata-get' advice.
+    (when-let (annotate (marginalia--orig-completion-metadata-get
+                         marginalia--metadata 'annotation-function))
       (funcall annotate cand))))
 
 (defconst marginalia--advice-regexp
@@ -482,6 +503,7 @@ l local (L modified compared to default value)
 & alias
 
 Other:
+G custom group
 a face
 t cl-type"
   (let ((class
@@ -515,7 +537,7 @@ t cl-type"
                                  (default-value s))))
                    '("L" . "local, modified from global")
                  '("l" . "local, unmodified")))
-             (if (custom-variable-p s)
+             (if (get s 'standard-value)
                  (if (ignore-errors
                        (not (equal (symbol-value s)
                                    (eval (car (get s 'standard-value))))))
@@ -526,6 +548,7 @@ t cl-type"
                   (cons "&" (format "alias for `%s'" (ignore-errors (indirect-variable s)))))
              (and (get s 'byte-obsolete-variable) '("-" . "obsolete"))))
           (list
+           (and (get s 'group-documentation) '("G" . "custom group"))
            (and (facep s) '("a" . "face"))
            (and (get s 'cl--class) '("t" . "cl-type")))))) ;; cl-find-class, cl--find-class
     (setq class (delq nil class))
@@ -570,10 +593,12 @@ t cl-type"
     (marginalia--fields
      (:left (marginalia-annotate-binding cand))
      ((marginalia--symbol-class sym) :face 'marginalia-type)
-     ((cond
-       ((fboundp sym) (marginalia--function-doc sym))
-       ((facep sym) (documentation-property sym 'face-documentation))
-       (t (documentation-property sym 'variable-documentation)))
+     ((if (fboundp sym) (marginalia--function-doc sym)
+        (cl-loop
+         for doc in '(variable-documentation
+                      face-documentation
+                      group-documentation)
+         thereis (ignore-errors (documentation-property sym doc))))
       :truncate 1.0 :face 'marginalia-documentation)
      ((abbreviate-file-name (or (symbol-file sym) ""))
       :truncate -0.5 :face 'marginalia-file-name))))
@@ -597,7 +622,7 @@ keybinding since CAND includes it."
   "Annotate imenu CAND with its documentation string."
   (when (derived-mode-p 'emacs-lisp-mode)
     ;; Strip until the last whitespace in order to support flat imenu
-    (marginalia-annotate-symbol (replace-regexp-in-string "^.* " "" cand))))
+    (marginalia-annotate-symbol (replace-regexp-in-string "\\`.* " "" cand))))
 
 (defun marginalia-annotate-function (cand)
   "Annotate function CAND with its documentation string."
@@ -623,7 +648,9 @@ keybinding since CAND includes it."
                     thereis (if (symbolp r)
                                 (eq r sym)
                               (string-match-p r name)))))
-    (propertize "*****" 'face 'marginalia-null))
+    (propertize "*****"
+                'face 'marginalia-null
+                'help-echo "Hidden due to `marginalia-censor-variables'"))
    (t
     (let ((val (symbol-value sym)))
       (pcase val
@@ -634,7 +661,9 @@ keybinding since CAND includes it."
         ((pred hash-table-p) (propertize "#<hash-table>" 'face 'marginalia-value))
         ((pred syntax-table-p) (propertize "#<syntax-table>" 'face 'marginalia-value))
         ;; Emacs bug#53988: abbrev-table-p throws an error
-        ((and (pred vectorp) (guard (ignore-errors (abbrev-table-p val))))
+        ((guard (static-if (< emacs-major-version 30)
+                    (and (vectorp val) (ignore-errors (abbrev-table-p val)))
+                  (abbrev-table-p val)))
          (propertize "#<abbrev-table>" 'face 'marginalia-value))
         ((pred char-table-p) (propertize "#<char-table>" 'face 'marginalia-value))
         ;; Emacs 29 comes with callable objects or object closures (OClosures)
@@ -643,9 +672,10 @@ keybinding since CAND includes it."
                  (and (fboundp 'oclosure-type) (oclosure-type val))))
         ((pred byte-code-function-p) (propertize "#<byte-code-function>" 'face 'marginalia-function))
         ((and (pred functionp) (pred symbolp))
-         ;; NOTE: We are not consistent here, values are generally printed unquoted. But we
-         ;; make an exception for function symbols to visually distinguish them from symbols.
-         ;; I am not entirely happy with this, but we should not add quotation to every type.
+         ;; We are not consistent here, values are generally printed
+         ;; unquoted. But we make an exception for function symbols to visually
+         ;; distinguish them from symbols.  I am not entirely happy with this,
+         ;; but we should not add quotation to every type.
          (format (propertize "#'%s" 'face 'marginalia-function) val))
         ((pred recordp) (format (propertize "#<record %s>" 'face 'marginalia-value) (type-of val)))
         ((pred symbolp) (propertize (symbol-name val) 'face 'marginalia-symbol))
@@ -741,6 +771,7 @@ keybinding since CAND includes it."
 (defun marginalia-annotate-minor-mode (cand)
   "Annotate minor-mode CAND with status and documentation string."
   (let* ((sym (intern-soft cand))
+         (message-log-max nil)
          (mode (if (and sym (boundp sym))
                    sym
                  (lookup-minor-mode-from-indicator cand)))
@@ -780,7 +811,7 @@ keybinding since CAND includes it."
 (defun marginalia--bookmark-type (bm)
   "Return bookmark type string of BM.
 The string is transformed according to `marginalia--bookmark-type-transforms'."
-  (let ((handler (or (bookmark-get-handler bm) 'bookmark-default-handler)))
+  (let ((handler (or (bookmark-prop-get bm 'handler) 'bookmark-default-handler)))
     (and
      ;; Some libraries use lambda handlers instead of symbols. For
      ;; example the function `xwidget-webkit-bookmark-make-record' is
@@ -800,16 +831,18 @@ The string is transformed according to `marginalia--bookmark-type-transforms'."
 (defun marginalia-annotate-bookmark (cand)
   "Annotate bookmark CAND with its file name and front context string."
   (when-let ((bm (assoc cand (bound-and-true-p bookmark-alist))))
-    (let ((front (bookmark-get-front-context-string bm)))
-      (marginalia--fields
-       ((marginalia--bookmark-type bm) :width 10 :face 'marginalia-type)
-       ((bookmark-get-filename bm)
-        :truncate -0.5 :face 'marginalia-file-name)
-       ((unless (or (not front) (equal front ""))
-          (concat (string-clean-whitespace
-                   (string-replace "\n" "\\n" front))
-                  (marginalia--ellipsis)))
-        :truncate -0.3 :face 'marginalia-documentation)))))
+    (marginalia--fields
+     ((marginalia--bookmark-type bm) :width 10 :face 'marginalia-type)
+     ((or (bookmark-prop-get bm 'filename)
+          (bookmark-prop-get bm 'location))
+      :truncate (if (bookmark-prop-get bm 'filename) -0.5 0.5)
+      :face 'marginalia-file-name)
+     ((let ((front (or (bookmark-prop-get bm 'front-context-string) ""))
+            (rear (or (bookmark-prop-get bm 'rear-context-string) "")))
+        (unless (and (string-blank-p front) (string-blank-p rear))
+          (string-clean-whitespace
+           (concat front (marginalia--ellipsis) rear))))
+      :truncate 0.5 :face 'marginalia-documentation))))
 
 (defun marginalia-annotate-customize-group (cand)
   "Annotate customization group CAND with its documentation string."
@@ -859,11 +892,13 @@ The string is transformed according to `marginalia--bookmark-type-transforms'."
 
 (defun marginalia-annotate-buffer (cand)
   "Annotate buffer CAND with modification status, file name and major mode."
-  (when-let (buffer (get-buffer cand))
-    (marginalia--fields
-     ((marginalia--buffer-status buffer))
-     ((marginalia--buffer-file buffer)
-      :truncate -0.5 :face 'marginalia-file-name))))
+  (when-let ((buffer (get-buffer cand)))
+    (if (buffer-live-p buffer)
+        (marginalia--fields
+         ((marginalia--buffer-status buffer))
+         ((marginalia--buffer-file buffer)
+          :truncate -0.5 :face 'marginalia-file-name))
+      (marginalia--fields ("(dead buffer)" :face 'error)))))
 
 (defun marginalia--full-candidate (cand)
   "Return completion candidate CAND in full.
@@ -881,47 +916,51 @@ component of a full file path."
     ;; necessary information (there's not much else we can do)
     cand))
 
-(defun marginalia--remote-protocol (path)
-  "Return the remote protocol of PATH."
+(defun marginalia--remote-file-p (file)
+  "Return non-nil if FILE is remote.
+The return value is a string describing the remote location,
+e.g., the protocol."
   (save-match-data
-    (setq path (substitute-in-file-name path))
-    (and (string-match "\\`/\\([^/|:]+\\):" path)
-         (match-string 1 path))))
+    (setq file (substitute-in-file-name file))
+    (cl-loop for r in marginalia-remote-file-regexps
+             if (string-match r file)
+             return (or (match-string 1 file) "remote"))))
 
 (defun marginalia--annotate-local-file (cand)
   "Annotate local file CAND."
-  (when-let (attrs (ignore-errors
-                     ;; may throw permission denied errors
-                     (file-attributes (substitute-in-file-name
-                                       (marginalia--full-candidate cand))
-                                      'integer)))
-    ;; HACK: Format differently accordingly to alignment, since the file owner
-    ;; is usually not displayed. Otherwise we will see an excessive amount of
-    ;; whitespace in front of the file permissions. Furthermore the alignment
-    ;; in `consult-buffer' will look ugly. TODO: Find a better solution!
-    (if (eq marginalia-align 'right)
+  (marginalia--in-minibuffer
+    (when-let (attrs (ignore-errors
+                       ;; may throw permission denied errors
+                       (file-attributes (substitute-in-file-name
+                                         (marginalia--full-candidate cand))
+                                        'integer)))
+      ;; HACK: Format differently accordingly to alignment, since the file owner
+      ;; is usually not displayed. Otherwise we will see an excessive amount of
+      ;; whitespace in front of the file permissions. Furthermore the alignment
+      ;; in `consult-buffer' will look ugly. Find a better solution!
+      (if (eq marginalia-align 'right)
+          (marginalia--fields
+           ;; File owner at the left
+           ((marginalia--file-owner attrs) :face 'marginalia-file-owner)
+           ((marginalia--file-modes attrs))
+           ((marginalia--file-size attrs) :face 'marginalia-size :width -7)
+           ((marginalia--time (file-attribute-modification-time attrs))
+            :face 'marginalia-date :width -12))
         (marginalia--fields
-         ;; File owner at the left
-         ((marginalia--file-owner attrs) :face 'marginalia-file-owner)
          ((marginalia--file-modes attrs))
          ((marginalia--file-size attrs) :face 'marginalia-size :width -7)
          ((marginalia--time (file-attribute-modification-time attrs))
-          :face 'marginalia-date :width -12))
-      (marginalia--fields
-       ((marginalia--file-modes attrs))
-       ((marginalia--file-size attrs) :face 'marginalia-size :width -7)
-       ((marginalia--time (file-attribute-modification-time attrs))
-        :face 'marginalia-date :width -12)
-       ;; File owner at the right
-       ((marginalia--file-owner attrs) :face 'marginalia-file-owner)))))
+          :face 'marginalia-date :width -12)
+         ;; File owner at the right
+         ((marginalia--file-owner attrs) :face 'marginalia-file-owner))))))
 
 (defun marginalia-annotate-file (cand)
   "Annotate file CAND with its size, modification time and other attributes.
 These annotations are skipped for remote paths."
-  (if-let (remote (or (marginalia--remote-protocol cand)
+  (if-let (remote (or (marginalia--remote-file-p cand)
                       (when-let (win (active-minibuffer-window))
                         (with-current-buffer (window-buffer win)
-                          (marginalia--remote-protocol (minibuffer-contents-no-properties))))))
+                          (marginalia--remote-file-p (minibuffer-contents-no-properties))))))
       (marginalia--fields (remote :format "*%s*" :face 'marginalia-documentation))
     (marginalia--annotate-local-file cand)))
 
@@ -1002,10 +1041,7 @@ These annotations are skipped for remote paths."
 (defvar-local marginalia--project-root 'unset)
 (defun marginalia--project-root ()
   "Return project root."
-  (with-current-buffer
-      (if-let (win (active-minibuffer-window))
-          (window-buffer win)
-        (current-buffer))
+  (marginalia--in-minibuffer
     (when (eq marginalia--project-root 'unset)
       (setq marginalia--project-root
             (or (let ((prompt (minibuffer-prompt))
@@ -1015,9 +1051,7 @@ These annotations are skipped for remote paths."
                         prompt)
                        (match-string 1 prompt)))
                 (when-let (proj (project-current))
-                  (cond
-                   ((fboundp 'project-root) (project-root proj))
-                   ((fboundp 'project-roots) (car (project-roots proj))))))))
+                  (project-root proj)))))
     marginalia--project-root))
 
 (defun marginalia-annotate-project-file (cand)
@@ -1030,17 +1064,15 @@ These annotations are skipped for remote paths."
 
 (defvar-local marginalia--library-cache nil)
 (defun marginalia--library-cache ()
-  "Return library cache hash table."
-  (with-current-buffer
-      (if-let (win (active-minibuffer-window))
-          (window-buffer win)
-        (current-buffer))
+  "Return hash table from library name to library file."
+  (marginalia--in-minibuffer
     ;; `locate-file' and `locate-library' are bottlenecks for the
     ;; annotator. Therefore we compute all the library paths first.
     (unless marginalia--library-cache
       (setq marginalia--library-cache (make-hash-table :test #'equal))
-      ;; Search in reverse because of shadowing
-      (dolist (dir (reverse load-path))
+      (dolist (dir (delete-dups
+                    (reverse ;; Reverse because of shadowing
+                     (append load-path (custom-theme--load-path))))) ;; Include themes
         (dolist (file (ignore-errors
                         (directory-files dir 'full
                                          "\\.el\\(?:\\.gz\\)?\\'")))
@@ -1083,6 +1115,10 @@ These annotations are skipped for remote paths."
       (put-text-property 0 1 'marginalia--library-doc doc file))
     doc))
 
+(defun marginalia-annotate-theme (cand)
+  "Annotate theme CAND with documentation and path."
+  (marginalia-annotate-library (concat cand "-theme")))
+
 (defun marginalia-annotate-library (cand)
   "Annotate library CAND with documentation and path."
   (setq cand (marginalia--library-name cand))
@@ -1108,11 +1144,11 @@ These annotations are skipped for remote paths."
     (let* ((tab (nth index tabs))
            (ws (alist-get 'ws tab))
            (bufs (window-state-buffers ws)))
-      ;; NOTE: When the buffer key is present in the window state
-      ;; it is added in front of the window buffer list and gets duplicated.
+      ;; When the buffer key is present in the window state it is added in front
+      ;; of the window buffer list and gets duplicated.
       (when (cadr (assq 'buffer ws)) (pop bufs))
       (marginalia--fields
-       (:left index :format " (%s)" :face 'marginalia-key)
+       (:left (1+ index) :format " (%s)" :face 'marginalia-key)
        ((if (eq (car tab) 'current-tab)
             (length (window-list nil 'no-minibuf))
           (length bufs))
@@ -1134,11 +1170,10 @@ These annotations are skipped for remote paths."
 
 (defun marginalia-classify-original-category ()
   "Return original category reported by completion metadata."
-  ;; NOTE: Use `alist-get' instead of `completion-metadata-get' to bypass our
-  ;; `marginalia--completion-metadata-get' advice!
-  (when-let (cat (alist-get 'category marginalia--metadata))
-    ;; Ignore Emacs 28 symbol-help category in order to ensure that the
-    ;; categories are refined to our categories function and variable.
+  ;; Bypass our `marginalia--completion-metadata-get' advice.
+  (when-let (cat (marginalia--orig-completion-metadata-get marginalia--metadata 'category))
+    ;; Ignore `symbol-help' category in order to ensure that the categories are
+    ;; refined to our categories function and variable.
     (and (not (eq cat 'symbol-help)) cat)))
 
 (defun marginalia-classify-symbol ()
@@ -1192,8 +1227,7 @@ completion UIs like Vertico or Icomplete."
    (when-let (align (text-property-any 0 (length ann) 'marginalia--align t ann))
      (setq marginalia--cand-width-max
            (max marginalia--cand-width-max
-                (* (ceiling (+ (string-width cand)
-                               (compat-call string-width ann 0 align))
+                (* (ceiling (+ (string-width cand) (string-width ann 0 align))
                             marginalia--cand-width-step)
                    marginalia--cand-width-step)))))
   (cl-loop
@@ -1207,17 +1241,19 @@ completion UIs like Vertico or Icomplete."
                    ('center `(+ center ,marginalia-align-offset))
                    ('left `(+ left ,(+ marginalia-align-offset marginalia--cand-width-max)))
                    ('right `(+ right ,(+ marginalia-align-offset 1
-                                         (- (compat-call string-width ann 0 align)
+                                         (- (string-width ann 0 align)
                                             (string-width ann)))))))
         ann))
      (list cand "" ann))))
 
 (defun marginalia--affixate (metadata annotator cands)
   "Affixate CANDS given METADATA and Marginalia ANNOTATOR."
-  ;; Compute minimum width of windows, which display the minibuffer.
-  ;; vertico-buffer displays the minibuffer in different windows. We may want
-  ;; to generalize this and detect other types of completion buffers, e.g.,
-  ;; Embark Collect or the default completion buffer.
+  ;; Compute minimum width of windows, which display the minibuffer, including
+  ;; the miniwindow.  In general the computed width corresponds to the full
+  ;; frame width, since the miniwindow spans the full frame.  For example
+  ;; `vertico-buffer' displays the minibuffer in a separate window.  Similarly,
+  ;; we could detect other types of completion buffers, e.g., Embark Collect or
+  ;; the default completion buffer, and compute smaller widths.
   (let* ((width (cl-loop for win in (get-buffer-window-list) minimize (window-width win)))
          (marginalia-field-width (min (/ width 2) marginalia-field-width))
          (marginalia--metadata metadata)
@@ -1265,7 +1301,7 @@ Remember `this-command' for `marginalia-classify-by-command-name'."
 
 (defun marginalia--base-position (completions)
   "Record the base position of COMPLETIONS."
-  ;; NOTE: As a small optimization we track the base position only for file
+  ;; As a small optimization we track the base position only for file
   ;; completions, since `marginalia--full-candidate' is currently used only by
   ;; the file annotation function.
   (when minibuffer-completing-file-name
@@ -1285,48 +1321,43 @@ Remember `this-command' for `marginalia-classify-by-command-name'."
         ;; Remember `this-command' in order to select the annotation function.
         (add-hook 'minibuffer-setup-hook #'marginalia--minibuffer-setup)
         ;; Replace the metadata function.
+        (advice-add (compat-function completion-metadata-get) :before-until #'marginalia--completion-metadata-get)
         (advice-add #'completion-metadata-get :before-until #'marginalia--completion-metadata-get)
         ;; Record completion base position, for `marginalia--full-candidate'
         (advice-add #'completion-all-completions :filter-return #'marginalia--base-position))
     (advice-remove #'completion-all-completions #'marginalia--base-position)
+    (advice-remove (compat-function completion-metadata-get) #'marginalia--completion-metadata-get)
     (advice-remove #'completion-metadata-get #'marginalia--completion-metadata-get)
     (remove-hook 'minibuffer-setup-hook #'marginalia--minibuffer-setup)))
 
 ;;;###autoload
 (defun marginalia-cycle ()
   "Cycle between annotators in `marginalia-annotator-registry'."
+  ;; Only show `marginalia-cycle' in M-x in recursive minibuffers
+  (declare (completion (lambda (&rest _) (> (minibuffer-depth) 1))))
   (interactive)
-  (if-let ((win (active-minibuffer-window))
-           (buf (window-buffer win)))
-      (with-current-buffer buf
-        (let* ((pt (max 0 (- (point) (minibuffer-prompt-end))))
-               (metadata (completion-metadata (buffer-substring-no-properties
-                                               (minibuffer-prompt-end)
-                                               (+ (minibuffer-prompt-end) pt))
-                                              minibuffer-completion-table
-                                              minibuffer-completion-predicate))
-               (cat (or (completion-metadata-get metadata 'category)
-                        (user-error "Marginalia: Unknown completion category")))
-               (ann (or (assq cat marginalia-annotator-registry)
-                        (user-error "Marginalia: No annotators found for category `%s'" cat))))
-          (marginalia--cache-reset)
-          (setcdr ann (append (cddr ann) (list (cadr ann))))
-          ;; When the builtin annotator is selected and no builtin function is
-          ;; available, skip to the next annotator. Note that we cannot use
-          ;; `completion-metadata-get' to access the metadata since we must
-          ;; bypass the `marginalia--completion-metadata-get' advice.
-          (when (and (eq (cadr ann) 'builtin)
-                     (not (assq 'annotation-function metadata))
-                     (not (assq 'affixation-function metadata))
-                     (not (plist-get completion-extra-properties :annotation-function))
-                     (not (plist-get completion-extra-properties :affixation-function)))
-            (setcdr ann (append (cddr ann) (list (cadr ann)))))
-          (message "Marginalia: Use annotator `%s' for category `%s'" (cadr ann) (car ann))))
-    (user-error "Marginalia: No active minibuffer")))
-
-;; Emacs 28: Only show `marginalia-cycle' in M-x in recursive minibuffers
-(put #'marginalia-cycle 'completion-predicate
-     (lambda (&rest _) (> (minibuffer-depth) 1)))
+  (with-current-buffer (window-buffer
+                        (or (active-minibuffer-window)
+                            (user-error "Marginalia: No active minibuffer")))
+    (let* ((end (minibuffer-prompt-end))
+           (pt (max 0 (- (point) end)))
+           (md (completion-metadata (buffer-substring-no-properties end (+ end pt))
+                                    minibuffer-completion-table
+                                    minibuffer-completion-predicate))
+           (cat (or (completion-metadata-get md 'category)
+                    (user-error "Marginalia: Unknown completion category")))
+           (ann (or (assq cat marginalia-annotator-registry)
+                    (user-error "Marginalia: No annotators found for category `%s'" cat))))
+      (marginalia--cache-reset)
+      (setcdr ann (append (cddr ann) (list (cadr ann))))
+      ;; When the builtin annotator is selected and no builtin function is
+      ;; available, skip to the next annotator. Bypass the
+      ;; `marginalia--completion-metadata-get' advice.
+      (when (and (eq (cadr ann) 'builtin)
+                 (not (marginalia--orig-completion-metadata-get md 'annotation-function))
+                 (not (marginalia--orig-completion-metadata-get md 'affixation-function)))
+        (setcdr ann (append (cddr ann) (list (cadr ann)))))
+      (message "Marginalia: Use annotator `%s' for category `%s'" (cadr ann) (car ann)))))
 
 (provide 'marginalia)
 ;;; marginalia.el ends here
